"""Facilities for saving and loading of experimental setups.

TODO:
- Could provide a simple interface to show the most recently saved files in
  a directory

:copyright: Copyright 2023-2024 by Matt Laporte.
:license: Apache 2.0. See LICENSE for details.
"""

from collections.abc import Callable
from datetime import datetime
import json
import logging
from pathlib import Path
from typing import Any, Optional

import jax.random as jr
from jaxtyping import PyTree
import equinox as eqx


from feedbax.misc import git_commit_id


logger = logging.getLogger(__name__)


def save(
    path: str | Path,
    tree: PyTree[eqx.Module],
    hyperparameters: Optional[dict] = None,
) -> None:
    """Save a PyTree to disk along with hyperparameters used to generate it.

    Assumes none of the hyperparameters are JAX arrays, as these are not
    JSON serialisable.

    Based on the Equinox serialisation [example](https://docs.kidger.site/equinox/examples/serialisation/).

    Arguments:
        path: The path of the file to be saved. Note that the file at this path
            will be overwritten if it exists.
        tree: The PyTree to save. Its structure should match the return
            type of a function `setup_func`, which will be passed to `load`.
        hyperparameters: A dictionary of arguments for
            `setup_func` that were used to generate the PyTree, and upon
            loading, will be used to regenerate an appropriate skeleton to
            populate with the saved values from `tree`.
    """
    with open(path, "wb") as f:
        hyperparameter_str = json.dumps(hyperparameters)
        f.write((hyperparameter_str + "\n").encode())
        eqx.tree_serialise_leaves(f, tree)


def load(
    path: Path | str,
    setup_func: Callable[..., PyTree[Any, "T"]],
) -> PyTree[Any, "T"]:
    """Setup a PyTree from stored data and hyperparameters.

    Arguments:
        path: The path of the file to be loaded.
        setup_func: A function that returns a PyTree of the same structure
            as the PyTree that was saved to `path`, and which may take as
            arguments `hyperparameters` which `save` may have saved to the same
            file. It must take a keyword argument `key`.
    """
    with open(path, "rb") as f:
        hyperparameters = json.loads(f.readline().decode())
        if hyperparameters is None:
            hyperparameters = dict()
        tree = setup_func(**hyperparameters, key=jr.PRNGKey(0))
        tree = eqx.tree_deserialise_leaves(f, tree)

    return tree


def _save_with_datetime_and_commit_id(
    tree: PyTree[eqx.Module],
    hyperparameters: Optional[dict] = None,
    path: Optional[str | Path] = None,
    save_dir: str | Path = Path("."),
    suffix: Optional[str] = None,
) -> None:
    """Save a PyTree to disk along with hyperparameters used to generate it.

    If a path is not specified, a filename will be generated from the current
    time and the commit ID of the `feedbax` repository, and the file will be
    saved in `save_dir`, which defaults to the current working directory.

    Assumes none of the hyperparameters are JAX arrays, as these are not
    JSON serializable.

    Based on https://docs.kidger.site/equinox/examples/serialisation/
    """

    if path is None:
        timestr = datetime.today().strftime("%Y%m%d-%H%M%S")
        commit_id = git_commit_id()
        name = f"{timestr}_{commit_id}"
        if suffix is not None:
            name += f"_{suffix}"
        path = Path(save_dir) / f"{name}.eqx"

    with open(path, "wb") as f:
        hyperparameter_str = json.dumps(hyperparameters)
        f.write((hyperparameter_str + "\n").encode())
        eqx.tree_serialise_leaves(f, tree)